/*
 * The MsgKernel program is an adaption of the core message-passing
 * mechanism from the SOS operating system. It demonstrates the use
 * of object concepts in expressing extensibility for messages and
 * modules.
 * @author Simon Han
 */
program MsgKernel {
    components { sched, Mica2, msg_pool, mod1, mod2 }
        entrypoint main = sched.entry;
}

class msg {
    field dest: int;
    field src:  int;
    field dispatch: function(msg_handler, msg): int;
    
    method send(dst: int, s: int, d: function(msg_handler, msg): int) : int {
        dest = dst;
        src = s;
        dispatch = d;
        return sched.msg_send(this);
    }
    method free() {
        msg_pool.free(this);
    }
}

component Dispatch {
    method init(h: msg_handler, m: msg): int { return h.handleInit(m); }
    method ping(h: msg_handler, m: msg): int { return h.handlePing(m); }
    method pong(h: msg_handler, m: msg): int { return h.handlePong(m); }
}

class msg_handler {
    field id: int;
    
    constructor(i: int) {
        id = i;
        sched.register(this);
    }

    method set_id(i: int) {
        id = i;
    }

    method get_id() : int {
        return id;
    }

    method handleInit(m: msg): int {
        return 0;
    }

    method handlePing(m: msg): int {
        return 0;
    }

    method handlePong(m: msg): int {
        return 0;
    }
}

component ID {
    field SCHED: int = 0;
    field TEST1: int = 1;
    field TEST2: int = 2;
}

component MSG_TYPE {
    field MSG_INIT: int = 0;
    field MSG_PING: int = 1;
    field MSG_PONG: int = 2;
}

class mod1 extends msg_handler {
    constructor() : super(ID.TEST1) { }

    method handleInit(m: msg): int {
        local m_out = msg_pool.alloc();
        Mica2.green.toggle();

        if(m_out != null) {
            m_out.send(ID.TEST2, ID.TEST1, Dispatch.ping);
        }
        m.free();
        return 0;
    }

    method handlePong(m: msg): int {
        Mica2.yellow.toggle();
        local m_out = msg_pool.alloc();
        if(m_out != null) {
            m_out.send(ID.TEST2, ID.TEST1, Dispatch.ping);
        }
        m.free();
        return 0;
    }
}


class mod2 extends msg_handler {
    constructor() : super(ID.TEST2) { }

    method handleInit(m: msg): int {
        Mica2.yellow.toggle();
        m.free();
        return 0;
        }
        method handlePing(m: msg): int {
        local m_out = msg_pool.alloc();
        Mica2.red.toggle();
        if(m_out != null) {
            m_out.send(m.src, ID.TEST2, Dispatch.pong);
        }
        m.free();
        return 0;
    }
}

component msg_pool {
    field NUM_MSG: int = 15;
    field pool: msg[] = new msg[NUM_MSG];
    
    constructor () {
        local i: int;
        for(i = 0; i < NUM_MSG; i++) {
            pool[i] = new msg();
        }
    }

    method alloc() : msg {
        local i: int;
        for(i = 0; i < NUM_MSG; i++) {
            if(pool[i] != null) {
                local tmp: msg = pool[i];
                pool[i] = null;
                return tmp;
            }
        }
        return null;
    }

    method free(m: msg) {
        local i: int;
        for(i = 0; i < NUM_MSG; i++) {
            if(pool[i] == null) {
                pool[i] = m;
                return;
            }
        }
    }
}

component sched {
    field MAX_MSG: int = 16;
    field MAX_HANDLE: int = 16;
    //! message handler registered to sched
    field hl: msg_handler[] = new msg_handler[MAX_HANDLE];
    //! message queue
    field mq: msg[]         = new msg[MAX_MSG];
    field mq_free: int;
    field mq_full: int;
    
    constructor () {
        local i: int;

        mq_free = 0;
        mq_full = 0;
        new mod1();
        new mod2();
    }

    method register(h: msg_handler): int {
        local i: int;
        for( i = 0; i < MAX_HANDLE; i ++ ) {
            if( hl[i] == null ) {
                local m: msg = msg_pool.alloc();
                hl[i] = h;
                
                if(m != null) {
                    m.send(h.id, ID.SCHED, Dispatch.init);
                }
                return 0;
            }
        }
        return -1;
    }

    method msg_send(m: msg): int {
        local tmp: int;
        tmp = mq_free;

        if( mq[tmp] == null ) {
            mq_free = (tmp + 1) % MAX_MSG;
            mq[tmp] = m;
            return 0;
        } else {
            return -1;
        }
    }

    method handle_msg(m: msg): int {
        local i: int;
        for( i = 0; i < MAX_HANDLE; i ++ ) {
            if( hl[i] != null and hl[i].get_id() == m.dest) {
                return m.dispatch(hl[i], m);
            }
        }
        return -1;
    }

    method entry() {
        Mica2.startLEDs();
        while( true ) {
            local old: int;
            local m: msg;
            old = mq_full;
            
            if( mq[old] != null ) {
                m = mq[old];
                mq[old] = null;
                mq_full = (old + 1) % MAX_MSG;

                // Handling message
                handle_msg( m );
            }
        }
    }
}

